/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package gramschmidt;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;
import javax.vecmath.Vector3d;

/**
 * Class that holds the data (vectors) used in the application.
 * Available data:
 * - initial vectors
 * - orthogonal vectors
 * - normalized orthogonal vectors
 * - projections used in calculations
 *
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public class GSDataModel {
    
    // initial vectors
    private Vector3d[] v;
    // orthogonal vectors
    private Vector3d[] u;
    // normalized orthogonal vectors
    private Vector3d[] e;
    // projections used in calculating the orthogonal vectors
    private Vector3d[] p;
    
    // listeners
    private EventListenerList ll;
    
    /**
     * Create a new data model.
     * Initialize this model before use.
     */
    public GSDataModel() {
        v = new Vector3d[3];
        u = new Vector3d[3];
        e = new Vector3d[3];
        p = new Vector3d[3];
        
        ll = new EventListenerList();
    }

    /**
     * Initialize the data using the given vectors.
     *
     * @return false if any of the vectors are null or of length 0. In this case, the data is unaltered.
     */
    public boolean initialize(Vector3d v1, Vector3d v2, Vector3d v3) {
        boolean accepted = v1 != null && v2 != null && v3 != null && v1.length() > 0 && v2.length() > 0 && v3.length() > 0;
        
        if (accepted) {
            v[0] = new Vector3d(v1);
            v[1] = new Vector3d(v2);
            v[2] = new Vector3d(v3);
            
            calculate();
            
            fireStateChanged(new ChangeEvent(this));
        }
        return accepted;
    }
    
    // perform Gram-Schmidt orthogonalization
    private void calculate() {
        u[0] = new Vector3d(v[0]);
        u[1] = new Vector3d(v[1]);
        p[0] = project(v[1],u[0]);
        u[1].sub(p[0]);
        u[2] = new Vector3d(v[2]);
        p[1] = project(v[2],u[0]);
        p[2] = project(v[2],u[1]);
        u[2].sub(p[1]);
        u[2].sub(p[2]);
        
        e[0] = new Vector3d(u[0]);
        e[0].normalize();
        e[1] = new Vector3d(u[1]);
        e[1].normalize();
        e[2] = new Vector3d(u[2]);
        e[2].normalize();
    }
    
    // returns a vector that is the orthogonal projection of v1 onto v2
    private Vector3d project(Vector3d v1, Vector3d v2) {
        Vector3d p = new Vector3d(v2);
        p.scale(v2.dot(v1) / v2.dot(v2));
        
        return p;
    }
    
    /**
     * @return initial vector i (i = 1,2,3).
     */
    public Vector3d getVector(int i) {
        if (i > 0 && i <= v.length)
            return new Vector3d(v[i - 1]);
        else
            return null;
    }
    
    /**
     * @return orthogonal vector i (i = 1,2,3).
     */
    public Vector3d getOrthogonalVector(int i) {
        if (i > 0 && i <= u.length)
            return new Vector3d(u[i - 1]);
        else
            return null;
    }
    
    /**
     * @return normalized orthogonal vector i (i = 1,2,3).
     */
    public Vector3d getNormalizedVector(int i) {
        if (i > 0 && i <= e.length)
            return new Vector3d(e[i - 1]);
        else
            return null;
    }
    
    /**
     * @return p(v2,u1), p(v3,u1) and p(v3,u2) for i = 1,2,3 respectively, where p(x,y) is the orthogonal projection of x onto y.
     */
    public Vector3d getProjectedVector(int i) {
        if (i > 0 && i <= p.length)
            return new Vector3d(p[i - 1]);
        else
            return null;
    }
    
    /**
     * Add a ChangeListener.
     */
    public void addChangeListener(ChangeListener l) {
        ll.add(ChangeListener.class, l);
    }
    
    /**
     * Remove a ChangeListener.
     */
    public void removeChangeListener(ChangeListener l) {
        ll.remove(ChangeListener.class, l);
    }
    
    /**
     * Notify all ChangeListeners.
     */
    public void fireStateChanged(ChangeEvent e) {
        ChangeListener[] listeners = ll.getListeners(ChangeListener.class);
        for (ChangeListener l : listeners)
            l.stateChanged(e);
    }
}