/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package gramschmidt;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import javax.swing.JPanel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

import static gramschmidt.GSStateModel.*;

/**
 * Panel that shows one step of the Gram-Schmidt orthogonalization process.
 * Data is provided by a GSDataModel and the step to draw is indicated by the GSStateModel.
 *
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public class GSPanel extends JPanel implements ChangeListener {
    
    // data
    private GSDataModel data;
    // state
    private GSStateModel state;
    
    // this indicates whether the initial vectors should be drawn, even if orthogonal replacements are already on screen
    private boolean showVectors = false;
    
    // color to use for the x/y/z axes
    private Color axesColor = new Color(0,0,0);
    // color to use for the initial vectors
    private Color vectorColor = new Color(0,0,128);
    // color to use for the orthogonal vectors
    private Color orthoColor = new Color(128,0,0);
    // color to use for the normalized vectors
    private Color normColor = new Color(0,128,0);
    // color to used for projections, subtractions and other aids
    private Color aidColor = new Color(64,64,64);
    
    // a dotted stroke
    private Stroke dottedStroke = new BasicStroke(1f, BasicStroke.CAP_SQUARE, BasicStroke.JOIN_MITER, 10.0f, new float[] {1, 3}, 0f);
    // a normal stroke
    private Stroke regularStroke = new BasicStroke();
    // a bold stroke
    private Stroke boldStroke = new BasicStroke(1.5f);
    
    /**
     * Create a new panel.
     * Don't forget to set a data and state model before use.
     */
    public GSPanel() {
        setPreferredSize(new Dimension(640,480));
        setBackground(Color.WHITE);
        
        PanelListener p = new PanelListener();
        addMouseListener(p);
        addMouseMotionListener(p);
        addMouseWheelListener(p);
    }
    
    /**
     * Set the data model to use.
     * If null is provided, the previous (current) data model is still used.
     */
    public void setDataModel(GSDataModel newData) {
        if (newData != null) {
            if (data != null)
                data.removeChangeListener(this);
            data = newData;
            data.addChangeListener(this);
            
            repaint();
        }
    }
    
    /**
     * Set the state model to use.
     * If null is provided, the previous (current) state model is still used.
     */
    public void setStateModel(GSStateModel newState) {
        if (newState != null) {
            if (state != null)
                state.removeChangeListener(this);
            state = newState;
            state.addChangeListener(this);
            
            repaint();
        }
    }
    
    /**
     * Set the option to show the initial vectors, even if orthogonal replacements are already on screen.
     * If set to false (which is the default), an initial vector will not be drawn once an orthogonalized version is available.
     */
    public void showVectors(boolean value) {
        showVectors = value;
        
        repaint();
    }
    
    // draw
    protected void paintComponent(Graphics g) {
        super.paintComponent(g);
        
        // all jaggies and no anti-aliasing makes this a dull drawing
        Graphics2D g2 = (Graphics2D) g;
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        
        // exit here if nothing is to be drawn
        if (state == null || data == null || state.getState() == STATE_EMPTY)
            return;
        
        int s = state.getState();
        
        // draw axes
        paintVector(new Vector3d(1,0,0), "x", g2, axesColor, dottedStroke);
        paintVector(new Vector3d(0,1,0), "y", g2, axesColor, dottedStroke);
        paintVector(new Vector3d(0,0,1), "z", g2, axesColor, dottedStroke);
        
        // draw initial vectors if they need to be drawn in the current state
        if (showVectors || s < STATE_ORTHOGONAL_1)
            paintVector(data.getVector(1), "v1", g2, vectorColor, regularStroke);
        if (showVectors || s < STATE_ORTHOGONAL_2)
            paintVector(data.getVector(2), "v2", g2, vectorColor, regularStroke);
        if (showVectors || s < STATE_ORTHOGONAL_3)
            paintVector(data.getVector(3), "v3", g2, vectorColor, regularStroke);
        
        // draw orthogonal vectors if they need to be drawn in the current state
        if (s >= STATE_ORTHOGONAL_1 && s <= STATE_ORTHOGONAL_3)
            paintVector(data.getOrthogonalVector(1), "u1", g2, orthoColor, boldStroke);
        if (s >= STATE_SUBTRACTION_1 && s <= STATE_ORTHOGONAL_3)
            paintVector(data.getOrthogonalVector(2), "u2", g2, orthoColor, boldStroke);
        if (s >= STATE_SUBTRACTION_2 && s <= STATE_ORTHOGONAL_3)
            paintVector(data.getOrthogonalVector(3), "u3", g2, orthoColor, boldStroke);
        
        // draw aids (projection lines, subtraction lines and markers to indicate the length of the projections)
        if (s == GSStateModel.STATE_PROJECTION_1) {
            paintLine(new Point3d(data.getVector(2)), new Point3d(data.getProjectedVector(1)), g2, aidColor, regularStroke);
            paintLine(new Point3d(0,0,0), new Point3d(data.getProjectedVector(1)), g2, aidColor, regularStroke);
            
            paintMarker(new Point3d(data.getProjectedVector(1)), g2, aidColor);
        } else if (s == GSStateModel.STATE_SUBTRACTION_1) {
            paintLine(new Point3d(data.getVector(2)), new Point3d(data.getProjectedVector(1)), g2, aidColor, regularStroke);
            paintLine(new Point3d(0,0,0), new Point3d(data.getProjectedVector(1)), g2, aidColor, regularStroke);
            paintLine(new Point3d(data.getVector(2)), new Point3d(data.getOrthogonalVector(2)), g2, aidColor, regularStroke);
            
            paintMarker(new Point3d(data.getProjectedVector(1)), g2, aidColor);
        } else if (s == GSStateModel.STATE_PROJECTION_2) {
            paintLine(new Point3d(data.getVector(3)), new Point3d(data.getProjectedVector(2)), g2, aidColor, regularStroke);
            paintLine(new Point3d(data.getVector(3)), new Point3d(data.getProjectedVector(3)), g2, aidColor, regularStroke);
            paintLine(new Point3d(0,0,0), new Point3d(data.getProjectedVector(2)), g2, aidColor, regularStroke);
            paintLine(new Point3d(0,0,0), new Point3d(data.getProjectedVector(3)), g2, aidColor, regularStroke);
            
            paintMarker(new Point3d(data.getProjectedVector(2)), g2, aidColor);
            paintMarker(new Point3d(data.getProjectedVector(3)), g2, aidColor);
        } else if (s == GSStateModel.STATE_SUBTRACTION_2) {
            paintLine(new Point3d(data.getVector(3)), new Point3d(data.getProjectedVector(2)), g2, aidColor, regularStroke);
            paintLine(new Point3d(data.getVector(3)), new Point3d(data.getProjectedVector(3)), g2, aidColor, regularStroke);
            paintLine(new Point3d(0,0,0), new Point3d(data.getProjectedVector(2)), g2, aidColor, regularStroke);
            paintLine(new Point3d(0,0,0), new Point3d(data.getProjectedVector(3)), g2, aidColor, regularStroke);
            
            Point3d v3mp2 = new Point3d(data.getVector(3));
            v3mp2.sub(data.getProjectedVector(2));
            Point3d v3mp3 = new Point3d(data.getVector(3));
            v3mp3.sub(data.getProjectedVector(3));
            
            paintLine(new Point3d(data.getVector(3)), v3mp2, g2, aidColor, regularStroke);
            paintLine(new Point3d(data.getVector(3)), v3mp3, g2, aidColor, regularStroke);
            paintLine(v3mp2, new Point3d(data.getOrthogonalVector(3)), g2, aidColor, regularStroke);
            paintLine(v3mp3, new Point3d(data.getOrthogonalVector(3)), g2, aidColor, regularStroke);
            
            paintMarker(new Point3d(data.getProjectedVector(2)), g2, aidColor);
            paintMarker(new Point3d(data.getProjectedVector(3)), g2, aidColor);
        } else if (s == GSStateModel.STATE_NORMALIZED) {
            // drawn normalized vectors
            paintVector(data.getNormalizedVector(1), "e1", g2, normColor, boldStroke);
            paintVector(data.getNormalizedVector(2), "e2", g2, normColor, boldStroke);
            paintVector(data.getNormalizedVector(3), "e3", g2, normColor, boldStroke);
        }
    }
    
    // draw a line between the given points, in color c and with stroke s
    private void paintLine(Point3d p1, Point3d p2, Graphics2D g2, Color c, Stroke s) {
        Point3d pp1 = projectPoint(p1);
        Point3d pp2 = projectPoint(p2);
        
        g2.setPaint(c);
        g2.setStroke(s);
        g2.draw(new Line2D.Double(pp1.x, pp1.y, pp2.x, pp2.y));
    }
    
    // draw a marker at point p in color c
    private void paintMarker(Point3d p, Graphics2D g2, Color c) {
        Point3d pp = projectPoint(p);
        
        g2.setPaint(c);
        g2.setStroke(new BasicStroke());
        g2.fill(new Ellipse2D.Double(pp.x - 2 , pp.y - 2, 4, 4));
        g2.draw(new Ellipse2D.Double(pp.x - 4 , pp.y - 4, 8, 8));
    }
    
    // draw the vector v with the given label, in color c and with stroke s
    private void paintVector(Vector3d v, String label, Graphics2D g2, Color c, Stroke s) {
        Point3d po = projectPoint(new Point3d(0,0,0));
        Point3d pp = projectPoint(new Point3d(v));
        
        g2.setPaint(c);
        g2.setStroke(s);
        g2.draw(new Line2D.Double(po.x, po.y, pp.x, pp.y));
        g2.fill(new Ellipse2D.Double(pp.x - 2 , pp.y - 2, 4, 4));
        g2.drawString(label, (int) pp.x + 3, (int) pp.y + 3);
    }
    
    // projection parameters
    private Vector3d screenPosition = new Vector3d(0, 0, 15);
    private Vector3d viewAngle = new Vector3d(235, 10, 180);
    private double modelScale = 10;
    
    // project the 3d input point onto the screen
    // this simple implementation is taken (with permission) from "Java 3D Programming" by Daniel Selman (www.manning.com/selman)
    private Point3d projectPoint(Point3d input) {
        double CT = Math.cos(Math.toRadians(viewAngle.x));
        double ST = Math.sin(Math.toRadians(viewAngle.x));
        double CP = Math.cos(Math.toRadians(viewAngle.y));
        double SP = Math.sin(Math.toRadians(viewAngle.y));
        
        double x = screenPosition.x + input.x * CT - input.y * ST;
        double y = screenPosition.y + input.x * ST * SP + input.y * CT * SP + input.z * CP;
        double temp = viewAngle.z / (screenPosition.z + input.x * ST * CP + input.y * CT * CP - input.z * SP);
        
        return new Point3d(getWidth() / 2 + modelScale * temp * x, getHeight() / 2 - modelScale * temp * y, 0);
    }
    
    // update the drawing
    public void stateChanged(ChangeEvent e) {
        repaint();
    }
    
    // handle mouse events
    private class PanelListener extends MouseAdapter implements MouseMotionListener, MouseWheelListener {
        private int x;
        private int y;
        private double vx;
        private double vy;
        
        // save the position where a mousedrag is started
        public void mousePressed(MouseEvent e) {
            x = e.getX();
            vx = viewAngle.x;
            y = e.getY();
            vy = viewAngle.y;
        }
        
        public void mouseMoved(MouseEvent e) {}
        
        // rotate left/right and up/down by changing the projection parameters appropriately
        public void mouseDragged(MouseEvent e) {
            int dx = e.getX() - x;
            int dy = e.getY() - y;
            
            viewAngle.x = vx + dx;;
            if (viewAngle.x < 0)
                viewAngle.x += 360;
            else if (viewAngle.x >= 360)
                viewAngle.x -= 360;
            
            viewAngle.y = vy + dy;
            if (viewAngle.y < 0)
                viewAngle.y += 360;
            else if (viewAngle.y >= 360)
                viewAngle.y -= 360;
            
            repaint();
        }
        
        // zoom in and out
        public void mouseWheelMoved(MouseWheelEvent e) {
            screenPosition.z += e.getWheelRotation();
            if (screenPosition.z < 1)
                screenPosition.z = 1;
            else if (screenPosition.z > 90)
                screenPosition.z = 90;
            
            repaint();
        }
    }
}
