/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package gramschmidt;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;

/**
 * State of the application (determines what is drawn).
 * Note that this class is not implemented as a state-machine since the states are simply sequential.
 *
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public class GSStateModel {
    
    // no data available
    public static final int STATE_EMPTY = 0;
    // initial vectors available
    public static final int STATE_INITIALIZED = 1;
    // one orthogonal vector available
    public static final int STATE_ORTHOGONAL_1 = 2;
    // show projection of v2 onto u1
    public static final int STATE_PROJECTION_1 = 3;
    // calculate u2 by subtraction
    public static final int STATE_SUBTRACTION_1 = 4;
    // two orthogonal vectors available
    public static final int STATE_ORTHOGONAL_2 = 5;
    // show projection of v3 onto u1 and u2
    public static final int STATE_PROJECTION_2 = 6;
    // calculate u3 by subtraction
    public static final int STATE_SUBTRACTION_2 = 7;
    // three orthogonal vectors available
    public static final int STATE_ORTHOGONAL_3 = 8;
    // normalize the vectors
    public static final int STATE_NORMALIZED = 9;
    
    // current state
    private int state;
    
    // listeners
    private EventListenerList ll;
    
    /**
     * Create a new state model.
     * This model will start at state STATE_EMPTY.
     */
    public GSStateModel() {
        state = STATE_EMPTY;
        
        ll = new EventListenerList();
    }
    
    /**
     * @return the current state.
     */
    public int getState() {
        return state;
    }
    
    /**
     * Set the current state.
     * Use one of the constants in GSStateModel as an argument.
     */
    public void setState(int state) {
        if (state >= STATE_EMPTY && state <= STATE_NORMALIZED) {
            this.state = state;
            fireStateChanged(new ChangeEvent(this));
        }
    }
    
    /**
     * @return the number of states that remain.
     * This is the difference between the final state and the current state.
     */
    public int statesRemaining() {
        return STATE_NORMALIZED - state;
    }
    
    /**
     * @return the number of states that are done.
     * This is the difference between the current state and the initial state (not the empty state).
     */
    public int statesDone() {
        return state > STATE_EMPTY ? state - 1 : 0;
    }
    
    /**
     * Move to the next state.
     * This method does nothing if the final state has been reached.
     */
    public void nextState() {
        if (statesRemaining() > 0) {
            state++;
            fireStateChanged(new ChangeEvent(this));
        }
    }
    
    /**
     * Move to the previous state.
     * This method does nothing if the initial state has been reached.
     */
    public void previousState() {
        if (statesDone() > 0) {
            state--;
            fireStateChanged(new ChangeEvent(this));
        }
    }
    
    /**
     * Add a ChangeListener.
     */
    public void addChangeListener(ChangeListener l) {
        ll.add(ChangeListener.class, l);
    }
    
    /**
     * Remove a ChangeListener.
     */
    public void removeChangeListener(ChangeListener l) {
        ll.remove(ChangeListener.class, l);
    }
    
    /**
     * Notify all ChangeListeners.
     */
    public void fireStateChanged(ChangeEvent e) {
        ChangeListener[] listeners = ll.getListeners(ChangeListener.class);
        for (ChangeListener l : listeners)
            l.stateChanged(e);
    }
}
